import React, {Component} from 'react'
import PropTypes from 'prop-types'
import { is, fromJS } from 'immutable'
import { notification } from 'antd'
import moment from 'moment'

import Api from '@/api'
import Utils from '@/utils/utils.js'
import UtilsDM from '@/utils/utils-datamanage.js'
import asyncComponent from '@/utils/asyncComponent'
import MKEmitter from '@/utils/events.js'
import MainTable from './normalTable'
import './index.scss'

// 通用组件
const MainSearch = asyncComponent(() => import('@/tabviews/zshare/topSearch'))
const MainAction = asyncComponent(() => import('@/tabviews/zshare/actionList'))
const NormalHeader = asyncComponent(() => import('@/tabviews/custom/components/share/normalheader'))

class EditableTable extends Component {
  static propTpyes = {
    config: PropTypes.object,        // 组件配置信息
  }

  state = {
    BID: '',              // 上级ID
    BData: '',            // 上级组件行数据
    config: {},           // 页面配置信息，包括按钮、搜索、显示列、标签等
    actions: null,        // 按钮集
    columns: null,        // 显示列
    setting: null,        // 页面全局设置：数据源、按钮及显示列固定、主键等
    data: [],             // 列表数据集
    selectedData: [],     // 已选表格数据
    total: 0,             // 总数
    loading: false,       // 列表数据加载中
    pageIndex: 1,         // 页码
    pageSize: 10,         // 每页数据条数
    orderBy: '',          // 排序
    search: '',           // 搜索条件数组，使用时需分场景处理
    allSearch: null
  }

  /**
   * @description 初始化处理
   * 1、 initdata 为打印时使用的数据集
   */
  UNSAFE_componentWillMount () {
    let _config = fromJS(this.props.config).toJS()
    let setting = {..._config.setting, ..._config.wrap}
    let BID = ''
    let BData = ''
    
    setting.tableId = Utils.getuuid().substr(-6).toUpperCase()

    if (_config.setting.supModule) {
      BData = window.GLOB.CacheData.get(_config.setting.supModule)
    } else {
      BData = window.GLOB.CacheData.get(_config.$pageId)
    }
    if (BData) {
      BID = BData.$BID || ''
    }

    _config.submit.style = _config.submit.style || {}
    _config.submit.wrapStyle = {}
    setting.hasAction = _config.action.length > 0

    _config.action.forEach(btn => {
      if (btn.funcType === 'addline') {
        setting.addable = true
      }
      btn.$tableId = setting.tableId
      if (!btn.controlField) {
        btn.controlField = '$lock'
        btn.controlVals = []
      }
    })

    let initId = ''

    setting.hasSubmit = false
    if (setting.commit === 'change' || setting.commit === 'simple') {
      setting.commit = 'change'
    }

    let index = 0
    let getColumns = (cols) => {
      return cols.filter(item => {
        if (item.Hide === 'true') return false

        item.tableId = setting.tableId

        if (item.type === 'colspan') {
          item.subcols = getColumns(item.subcols)

          if (item.subcols.length === 0) {
            return false
          }
        } else if (item.type === 'custom') {
          item.config = {setting: _config.setting, columns: _config.columns}

          item.elements.forEach(btn => {
            if (btn.eleType !== 'button') return
            if (btn.funcType === 'addline') {
              setting.addable = true
            }
            btn.$tableId = setting.tableId
            if (!btn.controlField) {
              btn.controlField = '$lock'
              btn.controlVals = []
            }
          })
        } else {
          if (item.editable === 'true') {
            item.$sort = index
            index++

            setting.hasSubmit = setting.commit !== 'change'
            item.$ctrl = setting.commit === 'change'
            
            if (item.ctrlField) {
              item.ctrlValue = item.ctrlValue.split(',')
            }
            if (!initId) {
              initId = item.uuid
            }
            if (item.enter === '$sub') {
              item.enter = '$noAct'
            } else if (item.enter === '$next') {
              item.enter = '$next_' + initId
            }

            if (item.linkSubField && item.linkSubField.length === 0) {
              item.linkSubField = null
            }

            if (item.type === 'number') {
              if (item.clearField) {
                _config.columns.forEach(cell => {
                  if (cell.field === item.clearField) {
                    item.clearName = cell.label
                  }
                })
              }
            } else if (item.editType === 'date') {
              item.format = 'YYYY-MM-DD'
  
              if (item.precision === 'hour') {
                item.format = 'YYYY-MM-DD HH'
              } else if (item.precision === 'minute') {
                item.format = 'YYYY-MM-DD HH:mm'
              } else if (item.precision === 'second') {
                item.format = 'YYYY-MM-DD HH:mm:ss'
              }
            } else if (item.editType === 'select') {
              item.map = new Map()

              item.options = item.options || []
              item.options = item.options.filter(cell => {
                cell.value = cell.Value
                cell.label = cell.Text
        
                return !cell.Hide
              })

              if (item.resourceType !== '1') {
                item.options.forEach(cell => {
                  item.map.set(cell.value, cell.label)
                })
              }
            } else if (item.editType === 'popSelect') {
              if (item.pops) {
                item.pops.forEach(cell => {
                  cell.linkSubField = item.linkSubField
                  this.resetPop(cell)
                  cell.tabName = cell.initval.replace(/^tab:/, '')
                })
              } else {
                this.resetPop(item)
              }
            }
          }
        }
  
        return true
      })
    }

    _config.cols = getColumns(_config.cols)

    let _columns = _config.cols
    if (_config.hasExtend) {
      _columns = this.getCols(_config.cols, BData, _config.setting.extendTime)
    }

    setting.initId = initId

    if (!_config.lineMarks || _config.lineMarks.length === 0) {
      _config.lineMarks = null
    }

    let allSearch = null
    if (_config.colsCtrls && _config.colsCtrls.length > 0) {
      allSearch = [{key: 'BID', value: BID || ''}]
      allSearch.push(..._config.$searches)
      if (_config.setting.useMSearch) {
        let mainSearch = window.GLOB.SearchBox.get(_config.$searchId) || []
        allSearch.push(...mainSearch)
      }
    } else {
      _config.colsCtrls = null
    }

    if (_config.setting.supModule && !BID) {
      _config.setting.onload = 'false'
    }

    this.setState({
      pageSize: setting.pageSize || 10,
      BID: BID || '',
      BData: BData || '',
      title: _config.wrap.title,
      config: _config,
      setting: setting,
      actions: _config.action,
      columns: _columns,
      search: _config.$searches,
      allSearch
    }, () => {
      if (_config.setting.onload === 'true') {
        setTimeout(() => {
          this.loadData()
        }, _config.setting.delay || 0)
      }
    })
  }

  resetPop = (config) => {
    let arrfield = config.columns.map(f => f.field)
  
    if (config.linkSubField && config.linkSubField.length > 0) {
      config.linkSubField.forEach(n => {
        if (!arrfield.includes(n)) {
          arrfield.push(n)
        }
      })
    }

    if (sessionStorage.getItem('dataM') === 'true') { // 数据权限
      config.dataSource = config.dataSource.replace(/\$@/ig, '/*').replace(/@\$/ig, '*/').replace(/@datam@/ig, '\'Y\'')
    } else {
      config.dataSource = config.dataSource.replace(/@\$|\$@/ig, '').replace(/@datam@/ig, '\'\'')
    }

    config.dataSource = config.dataSource.replace(/@LoginUID@/ig, `'${sessionStorage.getItem('LoginUID') || ''}'`)
    config.dataSource = config.dataSource.replace(/@SessionUid@/ig, `'${localStorage.getItem('SessionUid') || ''}'`)
    config.dataSource = config.dataSource.replace(/@UserID@/ig, `'${sessionStorage.getItem('UserID') || ''}'`)
    config.dataSource = config.dataSource.replace(/@Appkey@/ig, `'${window.GLOB.appkey || ''}'`)
    config.dataSource = config.dataSource.replace(/@lang@/ig, `'${sessionStorage.getItem('lang')}'`)

    if (/\s/.test(config.dataSource)) { // 拼接别名
      config.dataSource = '(' + config.dataSource + ') tb'
    }

    let labels = {}
    config.cols = []
    config.columns.forEach(col => {
      labels[col.field] = col.label

      if (col.Hide === 'true') return

      config.cols.push({
        dataIndex: col.field,
        title: col.label,
        sorter: col.IsSort === 'true',
        width: col.Width || 120
      })
    })

    let placeholder = ''
    if (!config.searchKey) {
      config.onload = 'true'
    } else {
      placeholder = []
      config.searchKey.split(',').forEach(key => {
        if (!labels[key]) {
          placeholder = ''
        } else if (placeholder) {
          placeholder.push(labels[key])
        }
      })

      placeholder = placeholder ? placeholder.join('、') : ''
    }

    config.placeholder = placeholder
    config.arr_field = arrfield.join(',')
  }

  getCols = (cols, BData, time) => {
    let _cols = []
    let weeks = ['星期日', '星期一', '星期二', '星期三', '星期四', '星期五', '星期六']

    let excols = (col) => {
      let _s = []
      let _marks = col.marks || []
      let names = []
      let fields = []

      for (let i = 0; i < col.quota; i++) {
        fields.push(col.field + (i === 0 ? '' : i))

        if (col.colUnit === 'day') {
          let hasweek = col.dayFormat.indexOf('week') > -1
          let format = col.dayFormat.replace(/\sweek/, '')
          let day = moment().add(col.shift + i, 'days').format(format)

          if (hasweek) {
            let _w = moment().add(col.shift + i, 'days').weekday()
            day = day + ' ' + weeks[_w]
          }

          names.push(day)
        } else {
          let format = col.hourFormat
          if (format === 'H point') {
            format = 'H点'
          }

          let hour = moment(time).add(col.shift + i, 'hours').format(format)

          if (format === 'h:00') {
            if (hour === '12:00') {
              hour = hour + ' pm'
            } else if (hour === moment(time).add(col.shift + i, 'hours').format('H:00')) {
              hour = hour + ' am'
            } else {
              hour = hour + ' pm'
            }
          }
          names.push(hour)
        }
      }

      if (col.supField) {
        names = []
        let val = ''
        if (BData) {
          let field = col.supField.toLowerCase()
          Object.keys(BData).forEach(key => {
            if (key.toLowerCase() === field) {
              val = BData[key] + ''
            }
          })
        }

        if (val) {
          names = val.split(',')

          if (names.length > fields.length) {
            names.length = fields.length
          } else if (fields.length > names.length) {
            fields.length = names.length
          }
        }
      }

      if (names.length === 0) return _s

      _s = fields.map((f, i) => {
        return {
          type: 'text',
          Align: col.Align,
          Width: col.Width,
          field: f,
          label: names[i],
          uuid: col.uuid + i,
          marks: _marks.filter(mark => mark.field[0] === f || !fields.includes(mark.field[0]))
        }
      })

      return _s
    }

    cols.forEach(col => {
      if (col.type === 'extend') {
        let exs = excols(col)
        _cols.push(...exs)
      } else if (col.type === 'colspan') {
        let subs = []
        col.subcols.forEach(subcol => {
          if (subcol.type === 'extend') {
            let exs = excols(subcol)
            subs.push(...exs)
          } else {
            subs.push(col)
          }
        })

        if (subs.length) {
          _cols.push({...col, subcols: subs})
        }
      } else {
        _cols.push(col)
      }
    })

    return _cols
  }

  resetCols = (BData) => {
    const { config, setting } = this.state

    MKEmitter.emit('transferData' + setting.tableId, [], 'clear')

    let columns = this.getCols(config.cols, BData, config.setting.extendTime)

    this.setState({
      columns: columns,
      data: [],
      selectedData: [],
      total: 0
    })
  }

  /**
   * @description 主表数据加载
   */
  async loadData (reset, repage) {
    const { setting, config, search, orderBy, BID, pageIndex, pageSize, BData } = this.state

    if (setting.supModule && !BID) { // BID 不存在时，不做查询
      this.setState({
        data: [],
        selectedData: [],
        loading: false,
        total: 0
      })
      reset && MKEmitter.emit('resetTable', config.uuid, 'true') // 列表重置
      MKEmitter.emit('transferData' + setting.tableId, [])

      this.requestId = ''
      return
    }

    let searches = fromJS(search).toJS()
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    if (config.$s_req && searches.filter(item => item.required && item.value === '').length > 0) {
      return
    }

    this.setState({
      loading: true
    })

    let _orderBy = orderBy || setting.order
    let param = UtilsDM.getQueryDataParams(setting, searches, _orderBy, pageIndex, pageSize, BID)

    this.requestId = config.uuid + new Date().getTime()

    let result = await Api.genericInterface(param, '', '', this.requestId)
    if (result.status) {
      if (result.$requestId && this.requestId !== result.$requestId) return
      
      reset && MKEmitter.emit('resetTable', config.uuid, repage) // 列表重置

      let start = 1
      if (setting.laypage) {
        start = pageSize * (pageIndex - 1) + 1
      }
      let data = result.data.map((item, index) => {
        item.key = index
        item.$$uuid = item[setting.primaryKey] || ''
        item.$$BID = BID || ''
        item.$$BData = BData || ''
        item.$Index = start + index + ''
        item.$type = 'upt'
        item.$origin = true

        if (config.absFields) {
          config.absFields.forEach(f => {
            if (!isNaN(item[f])) {
              item[f] = Math.abs(item[f])
            }
          })
        }

        return item
      })

      let total = result.total || 0
      if (config.setting.custompage && data.length) {
        total = data[data.length - 1].mk_total || 0
      }

      let allSearch = null
      if (config.colsCtrls) {
        allSearch = [{key: 'BID', value: BID || ''}]
        allSearch.push(...searches)
      }

      this.setState({
        data: data,
        selectedData: [],
        allSearch: allSearch,
        total: total,
        loading: false
      })

      MKEmitter.emit('transferData' + setting.tableId, data)

      UtilsDM.querySuccess(result)
    } else {
      this.setState({
        loading: false
      })
      
      UtilsDM.queryFail(result)
    }
  }

  /**
   * @description 获取单行数据
   */ 
  async loadLinedata (id, position) {
    const { setting, config, search, orderBy, BID, pageIndex, pageSize, BData } = this.state
    
    let searches = fromJS(search).toJS()
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    this.setState({
      loading: true
    })

    let _orderBy = orderBy || setting.order
    let param = UtilsDM.getQueryDataParams(setting, searches, _orderBy, pageIndex, pageSize, BID, id)

    let result = await Api.genericInterface(param)
    if (result.status) {
      if (position === 'line_grid' && (!result.data || !result.data[0])) {
        this.reloadIndex()
        return
      }

      if (!result.data || !result.data[0]) {
        let data = fromJS(this.state.data).toJS()
        let selectedData = fromJS(this.state.selectedData).toJS()

        data = data.filter(item => item.$$uuid !== id)
        selectedData = selectedData.filter(item => item.$$uuid !== id)

        MKEmitter.emit('transferData' + setting.tableId, id, 'delete')

        this.setState({
          data,
          selectedData,
          loading: false
        })
        return
      }

      let data = fromJS(this.state.data).toJS()
      let selectedData = fromJS(this.state.selectedData).toJS()

      let _data = result.data[0]
      _data.$$uuid = _data[setting.primaryKey] || ''
      _data.$$BID = BID || ''
      _data.$$BData = BData || ''
      _data.$type = 'upt'
      _data.$origin = true

      if (config.absFields) {
        config.absFields.forEach(f => {
          if (!isNaN(_data[f])) {
            _data[f] = Math.abs(_data[f])
          }
        })
      }

      data = data.map(item => {
        if (item.$$uuid === _data.$$uuid) {
          _data.key = item.key
          _data.$Index = item.$Index
          return _data
        } else {
          return item
        }
      })
      selectedData = selectedData.map(item => {
        if (_data.$$uuid === item.$$uuid) {
          return _data
        }
        return item
      })

      MKEmitter.emit('transferData' + setting.tableId, _data, 'line')
      MKEmitter.emit('resetSelectLine', config.uuid, _data.$$uuid, _data)

      this.setState({
        data,
        selectedData,
        loading: false
      })
    } else {
      this.setState({
        loading: false
      })
      
      UtilsDM.queryFail(result)
    }
  }

  /**
   * @description 搜索条件改变时，重置表格数据
   * 含有初始不加载的页面，修改设置
   */
  refreshbysearch = (searches) => {
    const { setting, BID } = this.state

    if (setting.supModule && !BID) {
      notification.warning({
        top: 92,
        message: window.GLOB.dict['sup_key_req'] || '需要上级主键值！',
        duration: 5
      })
      this.setState({
        search: searches
      })
      return
    }

    this.setState({
      pageIndex: 1,
      search: searches
    }, () => {
      this.loadData(true, 'true')
    })
  }

  /**
   * @description 表格条件改变时重置数据（分页或排序）
   */
  refreshbytable = (pagination, filters, sorter) => {
    if (sorter.order) {
      let _chg = {
        ascend: 'asc',
        descend: 'desc'
      }
      sorter.order = _chg[sorter.order]
    }

    this.setState({
      pageIndex: pagination.current,
      pageSize: pagination.pageSize,
      orderBy: (sorter.field && sorter.order) ? `${sorter.field} ${sorter.order}` : ''
    }, () => {
      this.loadData()
    })
  }

  /**
   * @description 表格刷新
   */
  reloadIndex = () => {
    this.setState({
      pageIndex: 1
    }, () => {
      this.loadData(true, 'true')
    })
  }

  /**
   * @description 导出Excel时，获取页面搜索排序等参数
   */
  queryModuleParam = (menuId, callback) => {
    const { config, orderBy, search, setting } = this.state

    if (config.uuid !== menuId) return

    let searches = search ? fromJS(search).toJS() : []
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    callback({
      orderBy: orderBy || setting.order,
      search: searches
    })
  }

  reloadData = (menuId, id) => {
    const { config } = this.state

    if (config.uuid !== menuId) return

    if (!id) {
      this.reloadIndex()
    } else if (config.forbidLine) {
      this.loadData(true, 'false')
    } else {
      this.loadLinedata(id)
    }
  }

  resetParentParam = (MenuID, id, data) => {
    const { setting, config } = this.state

    if (!setting.supModule || setting.supModule !== MenuID) return
    if (id !== this.state.BID || id !== '') {
      if (config.parCtrl) {
        this.resetCols(data)
      }

      this.setState({
        pageIndex: 1,
        BID: id,
        BData: data
      }, () => {
        if (!setting.checkBid) {
          setTimeout(() => {
            this.loadData(true, 'true')
          }, setting.delay || 0)
        }
      })
    }
  }

  /**
   * @description 按钮执行完成后页面刷新
   * @param {*} menuId     // 菜单Id
   * @param {*} position   // 刷新位置
   * @param {*} btn        // 执行的按钮
   */
  refreshByButtonResult = (menuId, position, btn, id, lines) => {
    const { config, BID } = this.state

    if (config.uuid !== menuId) return

    if (position === 'line' || position === 'line_grid') {
      if (lines && lines.length === 1) {
        if (config.forbidLine) {
          this.loadData(true, 'false')
        } else {
          this.loadLinedata(lines[0].$$uuid, position)
        }
      } else {
        this.loadData(true, 'false')
      }
    } else if (['mainline', 'maingrid', 'popclose'].includes(position) && config.setting.supModule) {
      MKEmitter.emit('reloadData', config.setting.supModule, position === 'maingrid' ? '' : BID)
    } else if (btn.resetPageIndex !== 'false') {
      this.reloadIndex()
    } else {
      this.loadData(true, 'false')
    }
  }

  refreshLineData = (menuId, btn, uuid, count) => {
    const { config } = this.state

    if (config.uuid !== menuId) return
    
    let _data = fromJS(this.state.data).toJS().map(item => {
      if (item.$$uuid === uuid) {
        item[btn.field] = count
      }
      return item
    })
    let _selectedData = fromJS(this.state.selectedData).toJS().map(item => {
      if (item.$$uuid === uuid) {
        item[btn.field] = count
      }
      return item
    })

    this.setState({
      data: _data,
      selectedData: _selectedData
    })
  }

  searchRefresh = (searchId) => {
    const { config } = this.state

    if (config.$searchId !== searchId) return
    
    this.reloadIndex()
  }

  shouldComponentUpdate (nextProps, nextState) {
    return !is(fromJS(this.props), fromJS(nextProps)) || !is(fromJS(this.state), fromJS(nextState))
  }

  componentDidMount () {
    const { config } = this.state

    MKEmitter.addListener('reloadData', this.reloadData)
    MKEmitter.addListener('resetSelectLine', this.resetParentParam)
    MKEmitter.addListener('queryModuleParam', this.queryModuleParam)
    MKEmitter.addListener('refreshLineData', this.refreshLineData)
    MKEmitter.addListener('refreshByButtonResult', this.refreshByButtonResult)

    if (config.setting.useMSearch) {
      MKEmitter.addListener('searchRefresh', this.searchRefresh)
    }
  }

  /**
   * @description 组件销毁，清除state更新，清除快捷键设置
   */
  componentWillUnmount () {
    this.setState = () => {
      return
    }
    MKEmitter.removeListener('reloadData', this.reloadData)
    MKEmitter.removeListener('searchRefresh', this.searchRefresh)
    MKEmitter.removeListener('refreshLineData', this.refreshLineData)
    MKEmitter.removeListener('resetSelectLine', this.resetParentParam)
    MKEmitter.removeListener('queryModuleParam', this.queryModuleParam)
    MKEmitter.removeListener('refreshByButtonResult', this.refreshByButtonResult)
  }

  render() {
    const { BID, setting, actions, config, columns, BData, data, selectedData, allSearch } = this.state

    let style = {...config.style}
    if (config.wrap.empty === 'hidden' && data.length === 0) {
      style.display = 'none'
    }

    return (
      <div className={'custom-edit-table' + (setting.hasSubmit ? '' : ' withnot-submit')} id={'anchor' + config.uuid} style={style}>
        <NormalHeader config={config}/>
        {config.search && config.search.length ?
          <MainSearch BID={BID} config={config} refreshdata={this.refreshbysearch}/> : null
        }
        {actions.length > 0 ? <MainAction
          BID={BID}
          setting={setting}
          actions={actions}
          BData={BData}
          columns={config.columns}
          selectedData={selectedData}
        /> : null}
        <MainTable
          BID={BID}
          setting={setting}
          columns={columns}
          MenuID={config.uuid}
          allSearch={allSearch}
          submit={config.submit}
          fields={config.columns}
          total={this.state.total}
          parCtrl={config.parCtrl}
          colsCtrls={config.colsCtrls}
          lineMarks={config.lineMarks}
          loading={this.state.loading}
          refreshdata={this.refreshbytable}
          chgSelectData={(selects) => this.setState({selectedData: selects})}
        />
      </div>
    )
  }
}

export default EditableTable